/*________________________________________________________________________________________________

	Title: Local Edge Intensity Analysis
	''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
	Authors: Jolyon Troscianko & Cedric van den Berg
	Date: 15/10/2018
............................................................................................................

This script measures the JNDs (based on the receptor noise-limited model and supplied Weber fractions)
of adjacent pixels in a cone-catch image. This is performaed at four different angles from horizontal,
diagonal, vertical and the other diagonal. The script measures the mean, stadnard deviaiton and skew
across the entire image. Zero-pixels are ignored (this is also a way of specifying a region to measure).

Unlike other boundary strength methods this does not require clustering, but should use an image
which has been acuity-controlled and having gone through noise-reduction.

The skew in boundary strength is likely to reflect the salience of an image's colour boundaries. This is
based on the adjusted Fisher–Pearson standardized moment coefficient.

--------------------------------------------------------------------------------------------------------------------------------
Requirements:

Input images should be cone-catch images. The last channel must be the luminance channel. e.g. a
trichromat human might be: LW, MW, SW, LUM and a tetrachromat bird might be: LW, MW, SW, UV, Double.


_________________________________________________________________________________________________
*/



import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.gui.*;
import ij.plugin.*;
import java.awt.*;

import ij.measure.ResultsTable;
import ij.text.TextWindow;
import ij.plugin.filter.Analyzer;
import ij.WindowManager;

import ij.plugin.frame.RoiManager;
import ij.gui.Roi; 



public class Local_Edge_Intensity_Analysis implements PlugInFilter {

ImageStack stack;
	public int setup(String arg, ImagePlus imp) { 
	stack = imp.getStack(); 
	//Roi roi = imp.getRoi();
	
	return DOES_32 + STACK_REQUIRED; 
	}



public void run(ImageProcessor ip) {


IJ.showStatus("JND Boundary Measurement");


int w = stack.getWidth();
int h = stack.getHeight();
int nPoints = w*h;
int nDims = stack.getSize();

String imTitle = "ImageID";

double[] weber = new double[nDims]; //

if(nDims == 4){
	weber[0] = 0.05; // LW HUMAN scaled to 0.05 for LWS from Hofer et al 2005, assuming a 2:1 L:M and 5.72% S cones
	weber[1] = 0.0707106781; // MW
	weber[2] = 0.1657433633; // SW
	weber[3] = 0.1; // 
} else {
for(int i = 0; i<nDims; i++)
	weber[i] = 0.05;

}

double lumWeber = 0.1;


String[] sliceNames = new String[nDims];
sliceNames = stack.getSliceLabels();
String[] transforms = new String[3];
transforms[0] = "none";
transforms[1] = "log";
transforms[2] = "sqrt";

	GenericDialog gd = new GenericDialog("JND Boundary Skew Analysis");

		gd.addStringField("Image ID:", imTitle, 20);
		gd.addMessage(" ");
		gd.addMessage("Weber Fractions");

		for(int i=0; i<nDims-1; i++)
			gd.addNumericField("Weber_" + (i+1) + " " + sliceNames[i], weber[i], 3);	
		
		gd.addNumericField("Luminance_Weber_Fraction", lumWeber, 3);
		gd.addChoice("Transform delta S values", transforms , transforms[0]);
		gd.addNumericField("Ignore_chromatic delta S values below", 0.000, 3);
		gd.addNumericField("Ignore_luminance delta S values below", 0.000, 3);

		gd.addCheckbox("Horizontal and vertical results", false);


	gd.showDialog();
	if (gd.wasCanceled())
		return;

imTitle = gd.getNextString();
for(int i=0; i<nDims-1; i++)
	weber[i] = gd.getNextNumber();
lumWeber = gd.getNextNumber();

int tChoice = gd.getNextChoiceIndex();
double ict = gd.getNextNumber();
double ilt = gd.getNextNumber();


boolean outputHVdata = gd.getNextBoolean();

//boolean outImage = gd.getNextBoolean();
	IJ.log("_____________________________________________");
	IJ.log("-----------Local Edge Intensity Analysis-----------");
	IJ.log("Label: " + imTitle);
	for(int i=0; i<nDims-1; i++)
		IJ.log("weber "+ sliceNames[i] + ": " + weber[i] );
	IJ.log("lumWeber: " + lumWeber);
	IJ.log("transform: " + transforms[tChoice]);
	IJ.log("chromatic delta S threshold: " + ict);
	IJ.log("luminance delta S threshold: " + ilt);


double[] dfis = new double[nDims];
double JND = 0.0;
double lumJND = 0.0;

ImageStack cJNDStack = new ImageStack(w, h);
ImageStack lJNDStack = new ImageStack(w, h);


double[] means = new double[nPoints*nDims];
int[] zeros = new int[nPoints];

float[] pixels= new float[w*h]; // temp array for loading

ImagePlus imp = IJ.getImage();
Roi roi = imp.getRoi();

if( roi == null ){ //-----------------No selection - Measuring whole image-----------------
	IJ.log("Measuring whole image");
	for(int k=0; k<w*h; k++)
		zeros[k] = 1;
} else { //-----------------Measuring ROI-----------------
	//IJ.log("Measuring Selection");
	ImageProcessor mask = roi.getMask();

	Rectangle r = roi.getBounds();

	if(mask == null){
		IJ.log("Measuring rectangular selection");
		for (int y=r.y; y<r.y+r.height; y++)
		for (int x=r.x; x<r.x+r.width; x++)
			zeros[(y*w) + x] = 1;
	} else {
		IJ.log("Measuring selected area");
		for (int y=r.y; y<r.y+r.height; y++)
		for (int x=r.x; x<r.x+r.width; x++)
		if(mask.getPixel(x-r.x,y-r.y) != 0)
			zeros[(y*w) + x] = 1;
	}


}
	IJ.log("_____________________________________________");


for(int i=0; i<nDims; i++){ // convert float pixel values to double-precision array
	pixels = (float[]) stack.getPixels(i+1);
	for(int k=0; k<w*h; k++)
		if(pixels[k] > 0) // don't import negative numbers
			means[(i*w*h)+k] = (double) (pixels[k]);
		else zeros[k] = 0;
}





//-----------------------------Initialise Results Table----------------------------

String[] titles = WindowManager.getNonImageTitles();
int sResultsOpen = 0;

for(int i=0; i<titles.length; i++){
	if(titles[i] == "Local Edge Intensity Analysis")
		sResultsOpen ++;
}

if(sResultsOpen > 0)
	IJ.renameResults("Local Edge Intensity Analysis", "Results");
	
ResultsTable st = Analyzer.getResultsTable();
if (st == null) {
	st = new ResultsTable();
	Analyzer.setResultsTable(st);
}

st.incrementCounter();
st.addValue("Image", imTitle);


//-----------------------------GENERATE JND EDGE MAPS (ON UNCLUSTERED IMAGE)---------------------------------
// Chromatic and achromatic maps of adjacent pixel JNDs, at 4 different angles

int uLoc = 0;
int vLoc = 0;
String angleLabel = "";

//--------sums of squares and sums of cubes-------------
double cJNDsum = 0.0;
double lJNDsum = 0.0;

double[] cJNDmeans = new double[4];
double[] lJNDmeans = new double[4];
double[] cJNDss = new double[4];
double[] lJNDss = new double[4];
double[] cJNDsc = new double[4];
double[] lJNDsc = new double[4];
double[] cJNDku = new double[4];
double[] lJNDku = new double[4];
//int nVals = (w-1)*(h-1);
//int nVals = 0;
double nVals = 0.0;

//------------Record maximum JND at each point-----
float[] cJNDmax = new float[nPoints];
float[] lJNDmax = new float[nPoints];

//		uLoc	vLoc
// horizontal		(y*w)+x	(y*w)+x+1
// diagonal \		(y*w)+x	((y+1)*w)+x+1
// vertical |		(y*w)+x	((y+1)*w)+x
// diagonal /		((y+1)*w)+x	(y*w)+x+1

for(int i=0; i<4; i++){

	float[] cJNDmap = new float[nPoints];
	float[] lJNDmap = new float[nPoints];

	cJNDsum = 0.0;
	lJNDsum = 0.0;

	//--------------------Calculate maps and means-------------------------------
	for(int y=0; y<h-1; y++)
	for(int x=0; x<w-1; x++){


		// ---------------Locations of adjacent pixels--------------------
		if(i==0){
			uLoc = (y*w)+x;
			vLoc = (y*w)+x+1;
		} else if(i==1){
			uLoc = (y*w)+x;
			vLoc = ((y+1)*w)+x+1;
		} else if(i==2){
			uLoc = (y*w)+x;
			vLoc = ((y+1)*w)+x;
		} else {
			uLoc = ((y+1)*w)+x;
			vLoc = (y*w)+x+1;
		}

		if(zeros[uLoc] == 1 && zeros[vLoc] == 1 ){

			for(int k=0; k<nDims-1; k++)
				dfis[k] = Math.log(means[(k*nPoints)+uLoc]/means[(k*nPoints)+vLoc]);

			if(nDims == 3) // DICHROMATIC EQUATION
				JND = Math.pow(Math.pow(dfis[0]-dfis[1],2)/(Math.pow(weber[0],2)+Math.pow(weber[1],2)),0.5);
			else if(nDims == 4) // TRICHROMATIC EQUATION
				JND = Math.pow((Math.pow(weber[0],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2],2)*Math.pow(dfis[0]-dfis[1],2))/(Math.pow(weber[0]*weber[1],2)+Math.pow(weber[0]*weber[2],2)+Math.pow(weber[1]*weber[2],2)),0.5);
			else if(nDims == 5) // TETRACHROMATIC EQUATION
				JND = Math.pow((Math.pow(weber[0]*weber[1],2)*Math.pow(dfis[3]-dfis[2],2)+Math.pow(weber[0]*weber[2],2)*Math.pow(dfis[3]-dfis[1],2)+Math.pow(weber[0]*weber[3],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1]*weber[2],2)*Math.pow(dfis[3]-dfis[0],2)+Math.pow(weber[1]*weber[3],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2]*weber[3],2)*Math.pow(dfis[1]-dfis[0],2))/(Math.pow(weber[0]*weber[1]*weber[2],2)+Math.pow(weber[0]*weber[1]*weber[3],2)+Math.pow(weber[0]*weber[2]*weber[3],2)+Math.pow(weber[1]*weber[2]*weber[3],2)),0.5);
			else {
				IJ.log("Image must have 2, 3 or 4 channels plus luminance in the last channel");
				return;
			}


			if(JND < ict)
				JND = 0.0;
			else if(tChoice == 1)
				JND = Math.log(JND);
			else if(tChoice == 2)
				JND = Math.pow(JND, 0.5);

			cJNDmap[uLoc] = (float) (JND);
			cJNDsum += JND;
			if(JND > cJNDmax[(y*w)+x])
				cJNDmax[(y*w)+x] = (float) (JND);

			// Siddiqi JNDs:
			lumJND=Math.abs((Math.log(means[((nDims-1)*nPoints)+uLoc])-Math.log(means[((nDims-1)*nPoints)+vLoc]))/lumWeber);

			if(lumJND < ilt)
				lumJND = 0.0;
			else if(tChoice == 1)
				lumJND = Math.log(lumJND);
			else if(tChoice == 2)
				lumJND = Math.pow(lumJND, 0.5);

			lJNDmap[uLoc] = (float) (lumJND);
			lJNDsum += lumJND;
			if(lumJND > lJNDmax[(y*w)+x])
				lJNDmax[(y*w)+x] = (float) (lumJND);

			if(i==0)
				nVals = nVals + 1; // count number of pairwise JNDs given the surrounding zeros
		} // non-zero comparison

	} // x y 

	if(i==0) angleLabel = "–";
	else if(i==1) angleLabel = "/";
	else if(i==2) angleLabel = "|";
	else  angleLabel = "\\";
	
	cJNDStack.addSlice(angleLabel, cJNDmap);
	lJNDStack.addSlice(angleLabel, lJNDmap);

	
	//--------------------Calculate SD, Skew and Kurtosis-------------------------------

	cJNDmeans[i] = cJNDsum / nVals;
	lJNDmeans[i] = lJNDsum / nVals;

	for(int y=0; y<h-1; y++)
	for(int x=0; x<w-1; x++){
		cJNDss[i] += Math.pow(	cJNDmap[(y*w)+x] - cJNDmeans[i] 	,2);
		lJNDss[i] += Math.pow(	lJNDmap[(y*w)+x] - lJNDmeans[i] 	,2);
		cJNDsc[i] += Math.pow(	cJNDmap[(y*w)+x] - cJNDmeans[i] 	,3);
		lJNDsc[i] += Math.pow(	lJNDmap[(y*w)+x] - lJNDmeans[i] 	,3);
		cJNDku[i] += Math.pow(	cJNDmap[(y*w)+x] - cJNDmeans[i] 	,4);
		lJNDku[i] += Math.pow(	lJNDmap[(y*w)+x] - lJNDmeans[i] 	,4);
	}


	// SDs
	cJNDss[i] = Math.pow(cJNDss[i]/(nVals-1),0.5);
	lJNDss[i] = Math.pow(lJNDss[i]/(nVals-1),0.5);

	//SKEWs --- based on adjusted Fisher–Pearson standardized moment coefficient
	cJNDsc[i] = (nVals * cJNDsc[i]) / (      (nVals-1)*(nVals-2)*Math.pow(cJNDss[i],3)    );
	lJNDsc[i] = (nVals * lJNDsc[i]) / (      (nVals-1)*(nVals-2)*Math.pow(lJNDss[i],3)    );

	//KURTOSIS --- based on adjusted Fisher–Pearson standardized moment coefficient
	cJNDku[i] = (nVals * cJNDku[i]) / (      (nVals-1)*(nVals-2)*Math.pow(cJNDss[i],4)    );
	lJNDku[i] = (nVals * lJNDku[i]) / (      (nVals-1)*(nVals-2)*Math.pow(lJNDss[i],4)    );

}//i


//---------------------Recalculate values for MAX------------------

double cJNDMaxsum = 0.0;
double lJNDMaxsum = 0.0;
double cJNDMaxss = 0.0;
double lJNDMaxss = 0.0;
double cJNDMaxsc = 0.0;
double lJNDMaxsc = 0.0;
double cJNDMaxku = 0.0;
double lJNDMaxku = 0.0;

for(int y=0; y<h-1; y++)
for(int x=0; x<w-1; x++){
	cJNDMaxsum += cJNDmax[(y*w)+x];
	lJNDMaxsum += lJNDmax[(y*w)+x];
}

//means
cJNDMaxsum = cJNDMaxsum/nVals;
lJNDMaxsum = lJNDMaxsum/nVals;

// sums-of-squares and cubes
for(int y=0; y<h-1; y++)
for(int x=0; x<w-1; x++){
	cJNDMaxss += Math.pow(	cJNDmax[(y*w)+x] - cJNDMaxsum	,2);
	lJNDMaxss += Math.pow(	lJNDmax[(y*w)+x] - lJNDMaxsum	,2);
	cJNDMaxsc += Math.pow(	cJNDmax[(y*w)+x] - cJNDMaxsum 	,3);
	lJNDMaxsc += Math.pow(	lJNDmax[(y*w)+x] - lJNDMaxsum 	,3);
	cJNDMaxku += Math.pow(	cJNDmax[(y*w)+x] - cJNDMaxsum 	,4);
	lJNDMaxku += Math.pow(	lJNDmax[(y*w)+x] - lJNDMaxsum 	,4);
}

// SDs
cJNDMaxss = Math.pow(cJNDMaxss/(nVals-1),0.5);
lJNDMaxss = Math.pow(lJNDMaxss/(nVals-1),0.5);

//SKEWs --- based on adjusted Fisher–Pearson standardized moment coefficient
cJNDMaxsc = (nVals * cJNDMaxsc) / (      (nVals-1)*(nVals-2)*Math.pow(cJNDMaxss,3)    );
lJNDMaxsc = (nVals * lJNDMaxsc) / (      (nVals-1)*(nVals-2)*Math.pow(lJNDMaxss,3)    );

//KURTOSIS --- based on adjusted Fisher–Pearson standardized moment coefficient
cJNDMaxku = (nVals * cJNDMaxku) / (      (nVals-1)*(nVals-2)*Math.pow(cJNDMaxss,4)    );
lJNDMaxku = (nVals * lJNDMaxku) / (      (nVals-1)*(nVals-2)*Math.pow(lJNDMaxss,4)    );



//---------------------------------------Output results of unclustered boundary strength analysis--------------------------------------------

st.addValue("Col mean", cJNDMaxsum);
st.addValue("Col sd", cJNDMaxss);
st.addValue("Col CoV", cJNDMaxss/cJNDMaxsum);
st.addValue("Col skew", cJNDMaxsc);
st.addValue("Col kurtosis", cJNDMaxku);

st.addValue("Lum mean", lJNDMaxsum);
st.addValue("Lum sd", lJNDMaxss);
st.addValue("Lum CoV", lJNDMaxss/lJNDMaxsum);
st.addValue("Lum skew", lJNDMaxsc);
st.addValue("Lum kurtosis", lJNDMaxku);

if(outputHVdata == true){
	st.addValue("Col mean hrz", cJNDmeans[0]);
	st.addValue("Col sd hrz", cJNDss[0]);
	st.addValue("Col CoV hrz", cJNDss[0]/cJNDmeans[0]);
	st.addValue("Col skew hrz", cJNDsc[0]);
	st.addValue("Col kurtosis hrz", cJNDku[0]);

	st.addValue("Col mean vrt", cJNDmeans[2]);
	st.addValue("Col sd vrt", cJNDss[2]);
	st.addValue("Col CoV vrt", cJNDss[2]/cJNDmeans[2]);
	st.addValue("Col skew vrt", cJNDsc[2]);
	st.addValue("Col kurtosis vrt", cJNDku[2]);

	st.addValue("Lum mean hrz", lJNDmeans[0]);
	st.addValue("Lum sd hrz", lJNDss[0]);
	st.addValue("Lum CoV hrz", lJNDss[0]/lJNDmeans[0]);
	st.addValue("Lum skew hrz", lJNDsc[0]);
	st.addValue("Lum kurtosis hrz", lJNDku[0]);

	st.addValue("Lum mean vrt", lJNDmeans[2]);
	st.addValue("Lum sd vrt", lJNDss[2]);
	st.addValue("Lum CoV vrt", lJNDss[2]/lJNDmeans[2]);
	st.addValue("Lum skew vrt", lJNDsc[2]);
	st.addValue("Lum kurtosis vrt", lJNDku[2]);
}


st.addValue("Transform", transforms[tChoice]);
st.addValue("Col thresh", ict);
st.addValue("Lum thresh", ilt);

st.showRowNumbers(false);
st.show("Results");

IJ.renameResults("Results", "Local Edge Intensity Analysis");


new ImagePlus(imTitle + "_Col_LEIA", cJNDStack).show();
new ImagePlus(imTitle + "_Lum_LEIA", lJNDStack).show();


}
}
